use 5.10.0;
use strict;
use Test::More tests => 1;

BEGIN { use_ok ('Geo::GoogleMaps::FitBoundsZoomer', 1.02); }

diag( "Testing Geo::GoogleMaps::FitBoundsZoomer $Geo::GoogleMaps::FitBoundsZoomer::VERSION, Perl $], $^X" );

